import { ClientBase, ClientConfig } from 'pg';
import { ServiceConfig } from './replication-service';
import { InboxMessage } from './models';
/** The inbox service configuration */
export type InboxServiceConfig = ServiceConfig & {
    /**
     * Database connection details. The user needs update permission to the inbox.
     */
    pgConfig: ClientConfig;
    settings: ServiceConfig['settings'] & {
        /**
         * The maximum number of attempts to handle an incoming inbox message.
         * Defaults to 5 which means a message is handled once initially and up to
         * four more times for retries.
         */
        maxAttempts?: number;
    };
};
/**
 * Message handler for a specific aggregate type and message type.
 */
export interface InboxMessageHandler {
    /** The aggregate root type */
    aggregateType: string;
    /** The name of the message created for the aggregate type. */
    messageType: string;
    /**
     * Custom business logic to handle a message that was stored in the inbox. It
     * is fine to throw an error if the message cannot be processed.
     * @param message The inbox message with the payload to handle.
     * @param client The database client that is part of a transaction to safely handle the inbox message.
     * @throws If something failed and the inbox message should NOT be acknowledged - throw an error.
     */
    handle: (message: InboxMessage, client: ClientBase) => Promise<void>;
    /**
     * Custom (optional) business logic to handle an error that was caused by the
     * "handle" method. You must ensure that this function does not throw an error
     * as the attempts counter is increased in the same transaction.
     * @param error The error that was thrown in the handle method.
     * @param message The inbox message with the payload that was attempted to be handled.
     * @param client The database client that is part of a (new) transaction to safely handle the error.
     * @param attempts The number of times that message was already attempted (including the current attempt). And the maximum number of times a message will be attempted.
     * @returns A flag that defines if the message should be retried ('transient_error') or not ('permanent_error')
     */
    handleError?: (error: Error, message: InboxMessage, client: ClientBase, attempts: {
        current: number;
        max: number;
    }) => Promise<ErrorType>;
}
export type ErrorType = 'transient_error' | 'permanent_error';
/**
 * Initialize the service to watch for inbox table inserts.
 * @param config The configuration object with required values to connect to the WAL.
 * @param messageHandlers A list of message handlers to handle the inbox messages. I
 * @returns Functions for a clean shutdown and to help testing "outages" of the inbox service
 */
export declare const initializeInboxService: (config: InboxServiceConfig, messageHandlers: InboxMessageHandler[]) => [shutdown: () => Promise<void>];
//# sourceMappingURL=inbox-service.d.ts.map