"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.initializeInboxService = void 0;
const pg_1 = require("pg");
const inbox_1 = require("./inbox");
const replication_service_1 = require("./replication-service");
const logger_1 = require("./logger");
const utils_1 = require("./utils");
/**
 * Initialize the service to watch for inbox table inserts.
 * @param config The configuration object with required values to connect to the WAL.
 * @param messageHandlers A list of message handlers to handle the inbox messages. I
 * @returns Functions for a clean shutdown and to help testing "outages" of the inbox service
 */
const initializeInboxService = (config, messageHandlers) => {
    const pool = createPgPool(config);
    const messageHandlerDict = getMessageHandlerDict(messageHandlers);
    const messageHandler = createMessageHandler(messageHandlerDict, pool, config);
    const errorResolver = createErrorResolver(messageHandlerDict, pool, config);
    const [shutdown] = (0, replication_service_1.createService)(config, messageHandler, errorResolver, mapInboxAttempts);
    return [
        () => __awaiter(void 0, void 0, void 0, function* () {
            pool.removeAllListeners();
            try {
                yield Promise.all([pool.end(), shutdown()]);
            }
            catch (e) {
                (0, logger_1.logger)().error(e, 'Inbox service shutdown error');
            }
        }),
    ];
};
exports.initializeInboxService = initializeInboxService;
const createPgPool = (config) => {
    const pool = new pg_1.Pool(config.pgConfig);
    pool.on('error', (err) => {
        (0, logger_1.logger)().error(err, 'PostgreSQL pool error');
    });
    return pool;
};
/**
 * Executes the inbox verification, the actual message handler, and marks the
 * inbox message as processed in one transaction.
 */
const createMessageHandler = (messageHandlers, pool, config) => {
    return (message) => __awaiter(void 0, void 0, void 0, function* () {
        yield (0, utils_1.executeTransaction)(pool, (client) => __awaiter(void 0, void 0, void 0, function* () {
            const result = yield (0, inbox_1.verifyInbox)(message, client, config);
            if (result === true) {
                const handler = messageHandlers[getMessageHandlerKey(message)];
                if (handler) {
                    yield handler.handle(message, client);
                }
                yield (0, inbox_1.ackInbox)(message, client, config);
            }
            else {
                (0, logger_1.logger)().warn(message, `Received inbox message cannot be processed: ${result}`);
            }
        }));
    });
};
const getMessageHandlerKey = (h) => `${h.aggregateType}@${h.messageType}`;
const getMessageHandlerDict = (messageHandlers) => {
    const handlers = {};
    for (const handler of messageHandlers) {
        const key = getMessageHandlerKey(handler);
        if (key in handlers) {
            throw new Error(`Only one message handler can handle one aggregate and message type. Multiple message handlers try to handle the aggregate type "${handler.aggregateType}" with the message type "${handler.messageType}".`);
        }
        handlers[key] = handler;
    }
    if (Object.keys(handlers).length === 0) {
        throw new Error('At least one message handler must be provided.');
    }
    return handlers;
};
/**
 * Increase the attempts counter of the message and retries it later.
 */
const createErrorResolver = (messageHandlers, pool, config) => {
    /**
     * An error handler that will increase the inbox attempts counter on transient errors.
     * @param message: the InboxMessage that failed to be processed
     * @param error: the error that was thrown while processing the message
     */
    return (message, error) => __awaiter(void 0, void 0, void 0, function* () {
        const handler = messageHandlers[getMessageHandlerKey(message)];
        try {
            yield (0, utils_1.executeTransaction)(pool, (client) => __awaiter(void 0, void 0, void 0, function* () {
                let errorType = 'transient_error';
                const maxAttempts = (0, inbox_1.getMaxAttempts)(config.settings.maxAttempts);
                if (handler === null || handler === void 0 ? void 0 : handler.handleError) {
                    errorType = yield handler.handleError(error, message, client, {
                        current: message.attempts + 1,
                        max: maxAttempts,
                    });
                }
                const attempts = errorType === 'permanent_error' ? maxAttempts : undefined;
                yield (0, inbox_1.nackInbox)(message, client, config, attempts);
            }));
        }
        catch (error) {
            (0, logger_1.logger)().error(Object.assign(Object.assign({}, message), { err: error }), handler
                ? 'The error handling of the message failed. Please make sure that your error handling code does not throw an error!'
                : 'The error handling of the message failed.');
        }
    });
};
/** The local replication service maps by default only the outbox properties */
const mapInboxAttempts = (input) => {
    var _a;
    if ('attempts' in input &&
        typeof input.attempts === 'number' &&
        'processed_at' in input &&
        (input.processed_at == null || input.processed_at instanceof Date)) {
        return {
            attempts: input.attempts,
            processedAt: (_a = input.processed_at) === null || _a === void 0 ? void 0 : _a.toISOString(),
        };
    }
    return {};
};
//# sourceMappingURL=inbox-service.js.map