import { PoolClient } from 'pg';
import { InboxServiceConfig } from './inbox-service';
import { OutboxMessage, InboxMessage } from './models';
/**
 * Initialize the inbox message storage to store incoming messages in the inbox table.
 * @param config The configuration object that defines the values on how to connect to the database and general settings.
 * @returns The function to store the inbox message data to the database and the shutdown action.
 */
export declare const initializeInboxMessageStorage: (config: Pick<InboxServiceConfig, 'pgConfig' | 'settings'>) => [storeInboxMessage: (message: OutboxMessage) => Promise<void>, shutdown: () => Promise<void>];
/**
 * Make sure the inbox item was not and is not currently being worked on.
 * As the WAL inbox service does not run in the same transaction as the message
 * handler code there is a small chance that the handler code succeeds but the
 * WAL inbox message was not acknowledged. This takes care of such cases.
 * @param message The inbox message to check and later process.
 * @param client The database client. Must be part of the transaction where the message handling changes are done.
 * @param config The configuration settings that defines inbox database schema.
 * @throws Throws an error if the database row does not exist, is locked by some other transaction, or was already processed.
 */
export declare const verifyInbox: ({ id }: InboxMessage, client: PoolClient, { settings }: Pick<InboxServiceConfig, 'settings'>) => Promise<true | 'INBOX_MESSAGE_NOT_FOUND' | 'ALREADY_PROCESSED' | 'MAX_ATTEMPTS_EXCEEDED'>;
/**
 * Acknowledge that the inbox message was handled.
 * @param message The inbox message to acknowledge.
 * @param client The database client. Must be part of the transaction where the message handling changes are done.
 * @param config The configuration settings that defines inbox database schema.
 */
export declare const ackInbox: ({ id }: InboxMessage, client: PoolClient, { settings }: Pick<InboxServiceConfig, 'settings'>) => Promise<void>;
/**
 * Reject the message and do not acknowledge it. Check the return value if the
 * WAL message should be acknowledged or not as the logic uses retries.
 * @param message The inbox message to NOT acknowledge.
 * @param client The database client. Must be part of the transaction where the message handling changes are done.
 * @param config The configuration settings that defines inbox database schema.
 * @param attempts Optionally set the number of attempts to a specific value. Especially relevant for non-transient errors to directly set it to the maximum attempts value.
 */
export declare const nackInbox: ({ id }: InboxMessage, client: PoolClient, { settings }: Pick<InboxServiceConfig, 'settings'>, attempts?: number) => Promise<void>;
export declare const getMaxAttempts: (maxAttempts?: number) => number;
//# sourceMappingURL=inbox.d.ts.map