import { ClientConfig } from 'pg';
import { Pgoutput } from 'pg-logical-replication';
import { OutboxMessage } from './models';
export interface ServiceConfig {
    /**
     * The "pg" library based settings to initialize the PostgreSQL connection for
     * the logical replication service (with replication permissions)
     */
    pgReplicationConfig: ClientConfig;
    /** service specific configurations */
    settings: ServiceConfigSettings;
}
export interface ServiceConfigSettings {
    /** The database schema name where the table is located */
    dbSchema: string;
    /** The database table of the inbox/outbox */
    dbTable: string;
    /** The name of the used PostgreSQL publication */
    postgresPub: string;
    /** The name of the used PostgreSQL logical replication slot */
    postgresSlot: string;
    /** When there is a message processing error it restarts the logical replication subscription with a delay. This setting defines this delay in milliseconds. Default is 250ms. */
    restartDelay?: number;
    /** When the replication slot is in use e.g. by another service, this service will still continue to try to connect in case the other service stops. Delay is given in milliseconds, the default is 10s. */
    restartDelaySlotInUse?: number;
    /** Inbox message handlers or the outbox message sender that do not finish can block further messages from being processed/sent. The timeout (in milliseconds) ensures to continue with the next items. Default is 15s. */
    messageProcessingTimeout?: number;
}
/**
 * Initiate the outbox/inbox table to listen for WAL messages.
 * @param config The replication connection settings and general service settings
 * @param messageHandler The message handler that handles the outbox/inbox message
 * @param errorHandler A handler that can decide if the WAL message should be acknowledged (true) or not (restarts the logical replication service)
 * @param mapAdditionalRows The inbox table requires an additional row to be mapped to the inbox message
 * @returns A function to stop the service
 */
export declare const createService: <T extends OutboxMessage>({ pgReplicationConfig, settings }: ServiceConfig, messageHandler: (message: T) => Promise<void>, errorHandler?: ((message: T, err: Error) => Promise<void>) | undefined, mapAdditionalRows?: ((row: object) => Record<string, unknown>) | undefined) => [shutdown: {
    (): Promise<void>;
}];
/** Get and map the inbox/outbox message if the WAL log entry is such a message. Otherwise returns undefined. */
declare const getRelevantMessage: <T extends OutboxMessage>(log: Pgoutput.Message, { dbSchema, dbTable }: ServiceConfig['settings'], mapAdditionalRows?: ((row: object) => Record<string, unknown>) | undefined) => T | undefined;
/** Maps the WAL log entry to an outbox or inbox message */
declare const mapMessage: <T extends OutboxMessage>(input: unknown, mapAdditionalRows?: ((row: object) => Record<string, unknown>) | undefined) => T | undefined;
/**
 * This export is _only_ done for unit tests as the createService function is
 * otherwise very hard to unit test. Exports work only for jest tests!
 */
export declare const __only_for_unit_tests__: {
    getRelevantMessage?: typeof getRelevantMessage;
    mapMessage?: typeof mapMessage;
};
export {};
//# sourceMappingURL=replication-service.d.ts.map