"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __asyncValues = (this && this.__asyncValues) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator], i;
    return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
    function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
    function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.__only_for_unit_tests__ = exports.createService = void 0;
const events_1 = require("events");
const pg_1 = require("pg");
const async_mutex_1 = require("async-mutex");
const pg_logical_replication_1 = require("pg-logical-replication");
const logger_1 = require("./logger");
const utils_1 = require("./utils");
/**
 * Initiate the outbox/inbox table to listen for WAL messages.
 * @param config The replication connection settings and general service settings
 * @param messageHandler The message handler that handles the outbox/inbox message
 * @param errorHandler A handler that can decide if the WAL message should be acknowledged (true) or not (restarts the logical replication service)
 * @param mapAdditionalRows The inbox table requires an additional row to be mapped to the inbox message
 * @returns A function to stop the service
 */
const createService = ({ pgReplicationConfig, settings }, messageHandler, errorHandler, mapAdditionalRows) => {
    const plugin = new pg_logical_replication_1.PgoutputPlugin({
        protoVersion: 1,
        publicationNames: [settings.postgresPub],
    });
    let client;
    let restartTimeout;
    let stopped = false;
    // Run the service in a self restarting background event "loop" until it gets stopped
    (function start() {
        if (stopped) {
            return;
        }
        restartTimeout = undefined;
        client = new pg_1.Client(Object.assign(Object.assign({}, pgReplicationConfig), { replication: 'database', stop: false }));
        // On errors stop the DB client and reconnect from a clean state
        const applyRestart = (promise) => {
            promise
                .catch((e) => {
                const err = (0, utils_1.ensureError)(e);
                (0, logger_1.logger)().error(err);
                return err;
            })
                .then((err) => {
                stopClient(client);
                return err;
            })
                .then((err) => {
                if (!stopped && !restartTimeout) {
                    restartTimeout = setTimeout(start, getRestartTimeout(err, settings));
                }
            });
        };
        // Start background functions to connect the DB client and handle replication data
        applyRestart(subscribe(client, plugin, settings.postgresSlot));
        applyRestart(handleIncomingData(client, plugin, settings, messageHandler, errorHandler, mapAdditionalRows));
    })();
    return [
        () => __awaiter(void 0, void 0, void 0, function* () {
            (0, logger_1.logger)().debug('started cleanup');
            stopped = true;
            if (restartTimeout) {
                clearTimeout(restartTimeout);
            }
            yield stopClient(client);
        }),
    ];
};
exports.createService = createService;
/** Get and map the inbox/outbox message if the WAL log entry is such a message. Otherwise returns undefined. */
const getRelevantMessage = (log, { dbSchema, dbTable }, mapAdditionalRows) => log.tag === 'insert' &&
    log.relation.schema === dbSchema &&
    log.relation.name === dbTable
    ? mapMessage(log.new, mapAdditionalRows)
    : undefined;
/** Maps the WAL log entry to an outbox or inbox message */
const mapMessage = (input, mapAdditionalRows) => {
    if (typeof input !== 'object' || input === null) {
        return undefined;
    }
    if (!('id' in input) ||
        typeof input.id !== 'string' ||
        !('aggregate_type' in input) ||
        typeof input.aggregate_type !== 'string' ||
        !('aggregate_id' in input) ||
        typeof input.aggregate_id !== 'string' ||
        !('message_type' in input) ||
        typeof input.message_type !== 'string' ||
        !('created_at' in input) ||
        !(input.created_at instanceof Date) || // date
        !('payload' in input) ||
        !('metadata' in input)) {
        return undefined;
    }
    const additional = mapAdditionalRows === null || mapAdditionalRows === void 0 ? void 0 : mapAdditionalRows(input);
    const message = Object.assign({ id: input.id, aggregateType: input.aggregate_type, aggregateId: input.aggregate_id, messageType: input.message_type, payload: input.payload, metadata: input.metadata, createdAt: input.created_at.toISOString() }, additional);
    return message;
};
const subscribe = (client, plugin, slotName, uptoLsn) => __awaiter(void 0, void 0, void 0, function* () {
    const lastLsn = uptoLsn || '0/00000000';
    yield client.connect();
    client.on('error', (e) => {
        (0, logger_1.logger)().error(e, 'client error');
    });
    client.connection.on('error', (e) => {
        (0, logger_1.logger)().error(e, 'connection error');
    });
    client.connection.once('replicationStart', () => {
        (0, logger_1.logger)().trace('Replication started');
    });
    return plugin.start(client, slotName, lastLsn);
});
function handleInboxOutboxMessage(client, settings, messageHandler, errorHandler, mapAdditionalRows, xLogData, lsn) {
    return __awaiter(this, void 0, void 0, function* () {
        const message = getRelevantMessage(xLogData, settings, mapAdditionalRows);
        if (!message) {
            acknowledge(client, lsn); // acknowledge messages that the service cannot use
            return;
        }
        try {
            (0, logger_1.logger)().debug(message, `Received a WAL message for ${settings.dbSchema}.${settings.dbTable}`);
            yield messageHandler(message);
            acknowledge(client, lsn);
        }
        catch (e) {
            (0, logger_1.logger)().error(xLogData, 'Error handling and acknowledging message');
            const err = (0, utils_1.ensureError)(e);
            if (errorHandler) {
                yield errorHandler(message, err);
            }
            throw e;
        }
    });
}
const stopClient = (client) => __awaiter(void 0, void 0, void 0, function* () {
    var _a;
    if (!client) {
        return;
    }
    try {
        (_a = client.connection) === null || _a === void 0 ? void 0 : _a.removeAllListeners();
        client.removeAllListeners();
        yield (0, utils_1.awaitWithTimeout)(() => client.end(), 1000, `PostgreSQL client could not be stopped within a reasonable time frame.`);
    }
    catch (e) {
        (0, logger_1.logger)().warn(e, 'Stopping the PostgreSQL client gave an error.');
    }
});
const getRestartTimeout = (error, settings) => {
    var _a, _b;
    const err = error;
    if ((err === null || err === void 0 ? void 0 : err.code) === '55006' && (err === null || err === void 0 ? void 0 : err.routine) === 'ReplicationSlotAcquire') {
        return (_a = settings.restartDelaySlotInUse) !== null && _a !== void 0 ? _a : 10000;
    }
    else {
        return (_b = settings.restartDelay) !== null && _b !== void 0 ? _b : 250;
    }
};
// The acknowledge function is based on the https://github.com/kibae/pg-logical-replication library
const acknowledge = (client, lsn) => {
    if (!(client === null || client === void 0 ? void 0 : client.connection)) {
        (0, logger_1.logger)().warn(`Could not acknowledge message ${lsn} as the client connection was not open.`);
        return false;
    }
    const slice = lsn.split('/');
    let [upperWAL, lowerWAL] = [
        parseInt(slice[0], 16),
        parseInt(slice[1], 16),
    ];
    // Timestamp as microseconds since midnight 2000-01-01
    const now = Date.now() - 946080000000;
    const upperTimestamp = Math.floor(now / 4294967.296);
    const lowerTimestamp = Math.floor(now - upperTimestamp * 4294967.296);
    if (lowerWAL === 4294967295) {
        // [0xff, 0xff, 0xff, 0xff]
        upperWAL = upperWAL + 1;
        lowerWAL = 0;
    }
    else {
        lowerWAL = lowerWAL + 1;
    }
    const response = Buffer.alloc(34);
    response.fill(0x72); // 'r'
    // Last WAL Byte + 1 received and written to disk locally
    response.writeUInt32BE(upperWAL, 1);
    response.writeUInt32BE(lowerWAL, 5);
    // Last WAL Byte + 1 flushed to disk in the standby
    response.writeUInt32BE(upperWAL, 9);
    response.writeUInt32BE(lowerWAL, 13);
    // Last WAL Byte + 1 applied in the standby
    response.writeUInt32BE(upperWAL, 17);
    response.writeUInt32BE(lowerWAL, 21);
    // Timestamp as microseconds since midnight 2000-01-01
    response.writeUInt32BE(upperTimestamp, 25);
    response.writeUInt32BE(lowerTimestamp, 29);
    // If 1, requests server to respond immediately - can be used to verify connectivity
    response.writeInt8(0, 33);
    client.connection.sendCopyFromChunk(response);
    return true;
};
// The acknowledge function is based on the https://github.com/kibae/pg-logical-replication library
// but with an async loop and a mutex to guarantee sequential processing
const LOG_FLAG = 0x77; // 119 in base 10
const KEEP_ALIVE_FLAG = 0x6b; // 107 in base 10
const handleIncomingData = (client, plugin, settings, messageHandler, errorHandler, mapAdditionalRows) => __awaiter(void 0, void 0, void 0, function* () {
    var _b, e_1, _c, _d;
    var _e;
    if (!client.connection) {
        throw new Error('Client not connected.');
    }
    const mutex = (0, async_mutex_1.withTimeout)(new async_mutex_1.Mutex(), (_e = settings.messageProcessingTimeout) !== null && _e !== void 0 ? _e : 15000);
    try {
        for (var _f = true, _g = __asyncValues((0, events_1.on)(client.connection, 'copyData')), _h; _h = yield _g.next(), _b = _h.done, !_b; _f = true) {
            _d = _h.value;
            _f = false;
            const [data] = _d;
            const release = yield mutex.acquire();
            try {
                // The semaphore ensure the sequential message processing to guarantee the correct order
                const { chunk: buffer, } = data;
                if (buffer[0] !== LOG_FLAG && buffer[0] !== KEEP_ALIVE_FLAG) {
                    (0, logger_1.logger)().warn({ bufferStart: buffer[0] }, 'Unknown message');
                    return;
                }
                const lsn = buffer.readUInt32BE(1).toString(16).toUpperCase() +
                    '/' +
                    buffer.readUInt32BE(5).toString(16).toUpperCase();
                if (buffer[0] === LOG_FLAG) {
                    const xLogData = plugin.parse(buffer.subarray(25));
                    yield handleInboxOutboxMessage(client, settings, messageHandler, errorHandler, mapAdditionalRows, xLogData, lsn);
                }
                else if (buffer[0] === KEEP_ALIVE_FLAG) {
                    // Primary keep alive message
                    const shouldRespond = !!buffer.readInt8(17);
                    if (shouldRespond) {
                        acknowledge(client, lsn);
                    }
                }
            }
            finally {
                release();
            }
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (!_f && !_b && (_c = _g.return)) yield _c.call(_g);
        }
        finally { if (e_1) throw e_1.error; }
    }
});
/**
 * This export is _only_ done for unit tests as the createService function is
 * otherwise very hard to unit test. Exports work only for jest tests!
 */
exports.__only_for_unit_tests__ = {};
if (process.env.JEST_WORKER_ID) {
    exports.__only_for_unit_tests__.getRelevantMessage = getRelevantMessage;
    exports.__only_for_unit_tests__.mapMessage = mapMessage;
}
//# sourceMappingURL=replication-service.js.map